# -*- coding: utf-8 -*-
"""
brickv (Brick Viewer)
Copyright (C) 2019 Erik Fleckstein <erik@tinkerforge.com>
Copyright (C) 2019 Matthias Bolte <matthias@tinkerforge.com>

pyinstaller_utils.py: PyInstaller utilities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
"""

import os
import shutil
import sys
import subprocess
import json
import time
import pprint

import PyInstaller.config

def specialize_template(template_filename, destination_filename, replacements):
    lines = []
    replaced = set()

    with open(template_filename, 'r') as f:
        for line in f.readlines():
            for key in replacements:
                replaced_line = line.replace(key, replacements[key])

                if replaced_line != line:
                    replaced.add(key)

                line = replaced_line

            lines.append(line)

    if replaced != set(replacements.keys()):
        raise Exception('Not all replacements for {0} have been applied'.format(template_filename))

    os.makedirs(os.path.dirname(destination_filename), exist_ok=True)

    with open(destination_filename, 'w+') as f:
        f.writelines(lines)

def system(command, stdout=None):
    if subprocess.call(command, stdout=stdout) != 0:
        sys.exit(1)

def by_ext(exts):
    def fn(full_name):
        if os.path.isfile(full_name):
            _, ext = os.path.splitext(full_name)
            ext = ext[1:]

            return ext in exts

    return fn

def by_name(name):
    return lambda full_name: os.path.basename(full_name) == name

def win_sign(exe_path):
    system(["C:\\Program Files (x86)\\Windows Kits\\10\\bin\\x86\\signtool.exe",
            "sign",
            "/v",
            "/fd", "sha256",
            "/tr", "http://rfc3161timestamp.globalsign.com/advanced",
            "/td", "sha256",
            "/a",
            "/n", "Tinkerforge GmbH",
            exe_path])
    system(["C:\\Program Files (x86)\\Windows Kits\\10\\bin\\x86\\signtool.exe", "verify", "/v", "/pa", exe_path])

class PyinstallerUtils:
    def __init__(self, name_words):
        self.underscore_name = '_'.join(name_words)
        self.camel_case_name = ' '.join([w.title() for w in name_words])
        self.version = None

        self.root_path = os.getcwd()
        self.build_path = PyInstaller.config.CONF['workpath']
        self.dist_path = PyInstaller.config.CONF['distpath']

        build_data_base_path = ''

        for argument in sys.argv:
            if argument.startswith('--build-data-path='):
                build_data_base_path = argument.replace('--build-data-path=', '')
            elif argument.startswith('--version='):
                self.version = argument.replace('--version=', '')

        assert self.version != None

        self.underscore_version = self.version.replace('.', '_').replace('+', '_').replace('~', '_')
        self.build_data_path = os.path.normpath(os.path.join(build_data_base_path))

        self.windows = sys.platform == 'win32'
        self.macos = sys.platform == 'darwin'
        self.linux = not self.windows and not self.macos

        self.win_dll_path = 'C:\\Program Files (x86)\\Windows Kits\\10\\Redist\\ucrt\\DLLs\\x86'

        if self.windows:
            self.pathex = [self.root_path, self.win_dll_path]
        else:
            self.pathex = [self.root_path]

        if self.windows:
            self.icon = os.path.join(self.build_data_path, self.underscore_name + '-icon.ico')
        elif self.linux:
            self.icon = self.underscore_name + '-icon.png'
        else:
            self.icon = os.path.join(self.build_data_path, self.underscore_name + '-icon.icns')

        self.datas = []

    def path_rel_to_root(self, path):
        return path.replace('\\', '/').replace(self.root_path.replace('\\', '/') + '/', '')

    def collect_data(self, pred):
        print("Collecting data")

        result = []

        for dirpath, _directories, files in os.walk(self.root_path):
            for file_ in files:
                full_name = os.path.join(dirpath, file_)

                if pred(full_name):
                    path_rel_to_root = self.path_rel_to_root(full_name)
                    result.append((path_rel_to_root, path_rel_to_root, 'DATA'))

        return result

    def win_build_installer(self):
        install_commands = []

        for root, dirs, files in os.walk(self.dist_path, topdown=False):
            if root == 'nsis':
                continue

            install_commands.append('  SetOutPath "{0}"'.format(os.path.normpath(os.path.join('$INSTDIR', os.path.relpath(root, self.dist_path)))))

            for file_ in files:
                path = os.path.normpath(os.path.relpath(os.path.join(root, file_), self.dist_path))

                install_commands.append('  File "..\\{0}"'.format(path))
                install_commands.append('  FileWrite $0 "$INSTDIR\\{0}$\\r$\\n"'.format(path))

            for dir_ in dirs:
                path = os.path.normpath(os.path.relpath(os.path.join(root, dir_), self.dist_path))

                install_commands.append('  FileWrite $0 "$INSTDIR\\{0}$\\r$\\n"'.format(path))

        nsis_template_path = os.path.join(self.build_data_path, 'nsis', self.underscore_name + '_installer.nsi.template')
        nsis_path = os.path.join(self.dist_path, 'nsis', self.underscore_name + '.nsi')

        specialize_template(nsis_template_path, nsis_path,
                            {'<<VERSION>>': self.version,
                             '<<UNDERSCORE_VERSION>>': self.underscore_version,
                             '<<INSTALL_COMMANDS>>': '\n'.join(install_commands)})

        system(['C:\\Program Files (x86)\\NSIS\\makensis.exe', nsis_path])

        installer = '{}_windows_{}.exe'.format(self.underscore_name, self.underscore_version)
        installer_target_path = os.path.join(self.root_path, '..', installer)

        if os.path.exists(installer_target_path):
            os.remove(installer_target_path)

        shutil.move(os.path.join(self.dist_path, 'nsis', installer), installer_target_path)

        return os.path.join(self.root_path, '..', installer)

    def prepare(self, prepare_script_working_dir=None, prepare_script=None):
        print('removing old dist directory')

        if os.path.exists(self.dist_path):
            shutil.rmtree(self.dist_path)

        if prepare_script is not None:
            if prepare_script_working_dir is not None:
                os.chdir(prepare_script_working_dir)

            print('calling {}'.format(prepare_script))
            system([sys.executable, prepare_script], stdout=subprocess.DEVNULL)

            if prepare_script_working_dir is not None:
                os.chdir(self.root_path)

        self.datas = self.collect_data(by_ext(['bmp', 'jpg', 'png', 'svg', 'obj', 'mtl', 'frag', 'vert']))
        self.datas += self.collect_data(by_name('internal'))
        self.datas += self.collect_data(by_name('snapshot'))
        self.datas += self.collect_data(by_name('package_type'))
        self.datas += self.collect_data(by_name('CertChain.crt'))

    def strip_binaries(self, binaries, patterns):
        return [x for x in binaries if all(pattern not in x[0].lower() for pattern in patterns)]

    def post_generate(self):
        if self.windows:
            self.post_generate_windows()
        elif self.macos:
            self.post_generate_macos()

    def post_generate_windows(self):
        exe_path = os.path.join(self.dist_path, self.underscore_name + '.exe')

        if '--no-sign' not in sys.argv:
            win_sign(exe_path)
        else:
            print("skipping win_sign")

        installer_exe_path = self.win_build_installer()

        if '--no-sign' not in sys.argv:
            win_sign(installer_exe_path)
        else:
            print("skipping win_sign for installer")

    def post_generate_macos(self):
        build_data = os.path.join(self.build_data_path, '*')
        entitlements_path = os.path.join(self.build_data_path, '..', 'entitlements.plist')
        app_name = self.camel_case_name + '.app'
        app_path = os.path.join(self.dist_path, app_name)
        resources_path = os.path.join(self.dist_path, app_name, 'Contents', 'Resources')
        system(['bash', '-c', 'cp -R {} {}'.format(build_data.replace(" ", "\\ "), resources_path.replace(" ", "\\ "))])

        if '--no-sign' not in sys.argv:
            system(['codesign', '--deep', '--force', '--verify', '--verbose=1', '-o', 'runtime', '--entitlements', entitlements_path, '--sign', 'Developer ID Application: Tinkerforge GmbH (K39N76HTZ4)', app_path])
            system(['codesign', '--verify', '--deep', '--verbose=1', app_path])

            print('notarize app')
            system(['ditto', '-c', '-k', '--keepParent', app_path, app_path + '.zip'])
            output = subprocess.check_output(['xcrun', 'notarytool', 'submit', app_path + '.zip', '--output-format', 'json', '--keychain-profile', 'notary-tinkerforge'])
            notarization_submit = json.loads(output)

            try:
                request_id = notarization_submit['id']
            except:
                print('error: notarization output does not contain expected fields')
                pprint.pprint(output)
                pprint.pprint(notarization_submit)
                sys.exit(1)

            print('notarize request id', request_id)
            notarization_info = None
            notarization_status = None

            while True:
                try:
                    output = subprocess.check_output(['xcrun', 'notarytool', 'info', request_id, '--output-format', 'json', '--keychain-profile', 'notary-tinkerforge'])
                except subprocess.CalledProcessError as e:
                    print('warning: notarize query failed, retrying', e)
                    time.sleep(1)
                    continue

                notarization_info = json.loads(output)

                try:
                    notarization_status = notarization_info['status']
                except:
                    print('error: notarization output does not contain expected fields')
                    pprint.pprint(output)
                    pprint.pprint(notarization_info)
                    sys.exit(1)

                if notarization_status != 'In Progress':
                    break

                print('waiting for notarization to complete ...')
                time.sleep(10)

            print('notarization info', notarization_info)

            if notarization_status != 'Accepted':
                print('error: notarization failed')
                sys.exit(1)

            print('staple notarization ticket to app')
            system(['xcrun', 'stapler', 'staple', app_path])
        else:
            print("skipping codesign")

        print('building disk image')
        dmg_path = os.path.join(self.dist_path, '..', '{}_macos_{}.dmg'.format(self.underscore_name, self.underscore_version))

        if os.path.exists(dmg_path):
            os.remove(dmg_path)

        os.mkdir(os.path.join(self.dist_path, 'dmg'))
        shutil.move(app_path, os.path.join(self.dist_path, 'dmg'))
        system(['hdiutil', 'create', '-fs', 'HFS+', '-volname', '{}-{}'.format(self.camel_case_name.replace(" ", "-"), self.version), '-srcfolder', os.path.join(self.dist_path, 'dmg'), dmg_path])
